\ 3iShift.fs -- Shift Register application code based on 3-Instruction Forth for AVR Atmel 328 / Arduino Nano
\ Assad Ebrahim, assadebrahim2000@gmail.com, http://www.mathscitech.org/articles
\ 2021-04-10---Sat---20:27 started, 20:47 complete, 20min to verify behaviour in software using 3iForth once wiring was complete!

\ FAST implementations of shift-in/shift-out registers running on uC (3iForth.avr)
$1E constant CALL_SHIFTINBYTE
$27 constant CALL_SHIFTIN16
$28 constant CALL_SHIFTIN24
$1C constant CALL_SHIFTOUTBYTE

: shout-byte ( byte -- )   \ send byte over serial and use shift-out register
     xstore CALL_SHIFTOUTBYTE xcall ;


\ SHIFT OUT REGISTER (POSI) -- 74595 IC DIP16 -- shift out from controller to chip, parallel out from chip to load pins
\ Chip has two 8-bit registers in memory, a shift register that is filled by pulsing the clock with data bits, which are latched into a storage register.  Output enable transfers the signals from storage register to load.
\ Useful for driving e.g. arrays of LEDs.
\ The output pins Q_A (1) to Q_H (8) can be thought of as an 8-bit FIFO queue with Q_A as entry and Q_H as exit, i.e. after 8 clock cycles, the first bit presented on SER INPUT line (Data In) will reside in Q_H.
\ The programmer needs to decide if Q_A will be interpreted as MSB or LSB of the 8-bit data.
\   This is decided in how data is fed in by controller, and how the Q pins are wired to the load.
\      e.g. using rshift to extract the LSB of a byte and feed into SER INPUT makes Q_A MSB and Q_H LSB.
\      With cascaded shift out registers, Q_A of the first SR is MSB and Q_H of last SR is LSB.
\ Both clock and latch are triggered on the rising edge of a pulse
\ Note - arduino has shiftOut() as a built-in function, but this uses software (bit-banging).  SPI pins would give a hardware behaviour.
\ Note - arduino has millis() for fine grained delay

\ WIRING CONNECTIONS (chip -> Nano, color):
\ see screenshot from chip datasheet
\ IC10 -> nanoA1/15 - CLEAR  grey  ( OPTIONAL --- equivalent to send 0 byte and clock it through )
\ IC11 -> nano13 - CLK  orange
\ IC12 -> nanoA0/14 - LATCH  yellow
\ IC13 -> nano10 - OUTPUT ENABLE  white   ( OPTIONAL --- could just hold output pins always tied to storage reg )
\ IC14 -> nano11 - DATA green
\ IC9  -> cascade --- tie this to IC14 of next chip (DATA) to cascade
\ IC16 -> power
\ IC8  -> ground

\ FIRMWARE - Arduino Nano running 3iForth.avr

\ USAGE:
\ include 3iForth.fs
\ 1 open-com
\ include 3iShift.fs

\ init-shiftout
\ clr

\ V1 = BIT-BANGING IMPLEMENTATION IN SERIAL (SLOW due to serial latency between Forth and Nano).
\ V2 = (TODO) send serial from Forth to Nano and do the bit banging at the Nano level.

\ fundamental functionality
\ : pin-enable 10 ;  \ CS  enables output from storage register to loads
\ : pin-latch  14 ;  
\ : pin-clk    13 ;  \ SCK
\ : pin-data   11 ;  \ COPI
\ : pin-clear  15 ;  \ NOT NEEDED since   0 set-byte   achieves the same (but slower) -- now locked to high (vcc)


\ NANO PIN DEFINITIONS SET to work with 13-key piano keyboard with 13 DIRECT INPUT BUTTON connections to Arduino Nano (pins 2-15/A1)
14 constant pin-data    \ COPI  (A0) green
15 constant pin-clk     \ SCK   (A1) orange
: clk         pin-clk pulse-hi ;
16 constant pin-latch   \       (A2) yellow
: latch       pin-latch pulse-hi ;
17 constant pin-enable  \ CS    (A3) enables output from storage register to loads  white  (or set to GND/LOW for permanent enable)
: enable-output  pin-enable lo ;
: disable-output pin-enable hi ;
\ : clr  pin-clear pulse-lo latch ;  \ wipes shift register to 0.  latch then wipes storage register.

\ REGISTER DEFINITIONS STORING the NANO pins for shift-out register wiring
20 constant REG_PINDATA
21 constant REG_PINCLOCK
22 constant REG_PINLATCH
23 constant REG_PINENABLE
: set-shiftout-pins  
    pin-data    REG_PINDATA   setreg \ DATA
    pin-clk     REG_PINCLOCK  setreg \ CLK
    pin-latch   REG_PINLATCH  setreg \ LAT
    pin-enable  REG_PINENABLE setreg \ CS (out enable), chip select,or peripheral select
;

\ manual bit-banging one clock pulse at a time.  Default pin-data state is low.  Latch to see.
: sh-hi  pin-data hi  clk  pin-data lo ;
: sh-lo  clk ;
: set-bit    ( bit -- ) pin-data pinwrite clk ;   \ sets data bit.  Issue is that this remembers state.
: [set-byte] ( byte --- ) 8 0 do dup 1 and set-bit  1 rshift loop  drop ;
: set-byte   ( byte --- ) [set-byte] latch ;
: clr ( -- ) 0 set-byte ;
: [set-id] ( id --- ) 1 swap 1 - lshift [set-byte] ;  \ count id from top of output queue (Q_H)
: set-id ( id --- ) [set-id] latch ;  \ count id from top of output queue (Q_H)
: set-id16 ( id -- ) disable-output 
                      dup 8 > if 0 [set-byte] 8 - set-id 
                              else [set-id] 0 set-byte 
                      then enable-output ;
: set-id32 ( id -- ) 
dup 8 <= if [set-id] 0 [set-byte] 0 [set-byte] 0 [set-byte] 
         else dup 16 <= if 0 [set-byte] 8 - [set-id] 0 [set-byte] 0 [set-byte]
                        else dup 24 <= if 0 [set-byte] 0 [set-byte] 16 - [set-id] 0 [set-byte]
                                       else 0 [set-byte] 0 [set-byte] 0 [set-byte] 24 -  [set-id]
                                       then
                        then
         then
         latch ;
\ 8-bit TESTS
: test-lights 256 0 do i set-byte loop ; \ show LEDs in binary
: l2r 8 1 do 8 i - set-id 1 ms loop ; 
: r2l 8 1 do i 1 + set-id 1 ms loop ; 
: kit begin l2r r2l key? until ;  \ knight rider style left/right.  Slow because of serial latency between Forth and Nano.


\ 2021-05-09---Sun---20:24  Adding Piano32 lights HW board with 4 cascaded shift out registers
: clear-32 disable-output 4 0 do 0 shout-byte loop enable-output ;  \ add/fix disable output-enable output
: on-32  4 0 do $FF shout-byte loop ; 
: c-shape clear-32 $89 shout-byte 0 set-bit ;  \ c + e + g in binary, LEDs 1 5 8 counting from left, or 1000 1001b = $89 = 137
: chf clk latch ; \ chord forward

variable ch-pos \ chord position
: chb  ; \ would need to have stored the starting position for the chord to move it back

: test-lights-shout
    256 0 do i shout-byte 15 ms loop 
    256 0 do 255 i - shout-byte loop ;
: test-32
 $2  shout-byte
 $6  shout-byte
 $E  shout-byte
 $1E shout-byte
;

: init-shiftout  \ sets output pins at defaults for 74595 shift out register IC
\    pin-clear out     pin-clear hi    \ ready for pulse-lo
    set-shiftout-pins
    pin-data   out      
    pin-clk    out       
    pin-latch  out     
    pin-enable out    
    pin-data lo     \ bit = 0
    pin-clk lo      \ ready for pulse-hi
    pin-latch lo    \ ready for pulse-hi
    disable-output
    \ $ff shout-byte 500 ms $0 shout-byte 500 ms
    clear-32
    enable-output
;


\ DEMONSTRATION CODES -----------------------
\ 8-bit ONE-PIP KIT
: shout-id ( id -- )  1 swap 1 - lshift shout-byte ;
: l2r 8 1 do 8 i - shout-id 50 ms loop ; 
: r2l 8 1 do i 1 + shout-id 50 ms loop ; 
: kit begin l2r 160 ms r2l 160 ms key? until ;  \ knight rider style left/right.  Can go super fast.
\ Exercise - could model the individual timing like pendulum, fastest in the middle then slowing at the ends
\ Exercise - send a two or 3 LED pip pulse there and back

\ TWO-PIP KIT
\ Knight rider style left/right. 
\ MUSIC: https://www.youtube.com/watch?v=x8bqWbys3GQ
\ Options: 1pip or 2pips
\ Timing: 
\   0 ms in l2r/r2l loops creates flashing of the outer 6 leds - imperceptible back and forth
\   10 ms  super fast - barely able to count laps
\   30 ms  fast - can now just about count lengths ** good if its only visual
\   70 ms  ** good if accompanied with audible whoosh-whoosh sound to go in time
\ Pause aesthetic:
\   kit1 is ricochet mode, no pause between sweeps
\   kit2 is sticky mode, slight pause between sweeps, light the knight rider car
30 constant wait
5 constant wavelet  \ binary wavelet that will be swept across, 3 is 2-pip, 5 is horned, 1 is 1-pip
: shout-id ( id -- )  wavelet swap 1 - lshift shout-byte ;  \ 3 in binary is the two-pip wavelet that we will sweep acrosss
: l2r 8 1 do 8 i - shout-id wait ms loop ; 
: r2l 7 1 do i 1 + shout-id wait ms loop ; 
: kit begin l2r r2l key? until 0 shout-byte ;  
: kit2 begin l2r wait 2 *  ms r2l wait ms key? until 0 shout-byte ;  
\ With pause at either end, one side is higher to comp to r2l loop less.
\ Exercise - could model the individual timing like pendulum, fastest in the middle then slowing at the ends

: 5-kit begin 5 1 do 5 i - shout-id 50 ms loop  90 ms 5 1 do i 1 + shout-id 50 ms loop 90 ms key? until 0 shout-byte ;

\ cascaded shift registers

: l2r16 16 1 do 16 i - set-id16 wait ms loop ; 
: r2l16 15 1 do i 1 + set-id16 wait ms loop ; 
: kit16 begin l2r16 wait 2 *  ms r2l16 wait ms key? until 0 shout-byte 0 shout-byte clear-32 ;
\ END DEMONSTRATION CODES -----------------------




\ ===============================================

\ SHIFT IN REGISTER IC 74165
\ Current draw: 15ma with all dip switches ON so going into IC and some leading through 10k resistors
\ Note - the ICs are drawing power directly from the clock, data, latch lines it looks like, i.e. the IC is sourcing parasitic power
\ from other pins --- do we need optoisolators to fully isolate from parasitic power?

\ Connections:
\ IC1 - LOAD/LATCH (yellow)
\ IC2 - CLOCK (orange)
\ IC9 -> DATA (green)

\ BANK A -- tones keypad - 16 keys
 6 constant pinA-dataIN    \ green 
 7 constant pinA-latch     \ yellow 
 8 constant pinA-clockIN   \ orange
\ BANK B -- config read - 24 bits
 3 constant pinB-dataIN   
 4 constant pinB-latch     
 5 constant pinB-clockIN
\ BANK C -- SPI keypad - 4x16 keys
 9 constant pinC-dataIN
10 constant pinC-latch
11 constant pinC-clockIN
12 constant pinC-keypressed

\ 13 constant speaker 
\ 13 led

: register-inpins  \ shiftin pins
    pinA-dataIN     28 setreg \ DATA
    pinA-latch      29 setreg \ LAT
    pinA-clockIN    30 setreg \ CLK
    pinB-dataIN     32 setreg \ DATA
    pinB-latch      33 setreg \ LAT
    pinB-clockIN    34 setreg \ CLK
    pinC-dataIN     39 setreg \ DATA
    pinC-latch      40 setreg \ LAT
    pinC-clockIN    41 setreg \ CLK
    pinC-keypressed 42 setreg \ CLK
;

\ CODE REVIEW:  move this into uC to make it more efficient
: init-shiftin
  register-inpins
  pinA-dataIN  in
  pinA-latch   out
  pinA-clockIN out
  pinB-dataIN  in
  pinB-latch   out
  pinB-clockIN out
  pinC-dataIN  in
  pinC-latch   out
  pinC-clockIN out
  pinC-keypressed in
  
  pinA-latch hi
  pinA-clockIN lo
  pinB-latch hi
  pinB-clockIN lo
  pinC-latch hi
  pinC-clockIN lo
;  

\ APPLICATION: BANK B,C - READING SPI HARDWARE MENU USING SLIDE SWITCHES AND 24-BITS SHIFTIN REGISTERS
: show8  ( b -- b )        cr dup . ." | MSB Q_8: " .b ." :Q_1 LSB |" ;
: show16  ( bhi blo -- bhi blo )      cr 2dup swap . . ." | MSB Q_16: " swap .b ." | Q_8 " swap .b ." :Q_1 LSB |" ;
: show24  ( bhi bmed blo -- bhi bmed blo ) cr 3 ndup drop  rot . swap . . 
        ." | MSB Q_24: " rot .b -rot ." | Q_16 " swap .b ." | Q_8 " swap .b ." :Q_1 LSB" ;

\ Design: shift-in registers work by loading 8 inputs in parallel using latch-in, and then extracting them bit by bit (serial)
\ Design: SLOW - reading bit or byte from shift-in register from host
\ WARNING: hardset to work with Bank A
: load-inputs pinA-latch lo ;
: hold-inputs pinA-latch hi ;
: latch-in ( -- ) load-inputs 1 ms hold-inputs ;
: shiftin-bit  ( -- bit ) pinA-dataIN pinread ( dup . ) pinA-clockIN pulse-hi  ;
\ MSB bits get read in and shifted left so last to be read is LSB
\ Q_H (dip 8) emitted first so this becomes MSB, Q_A (dip 1) is emitted last so this becomes LSB 
: shiftin-byte ( -- byte ) 8 0 do shiftin-bit loop >b8 ;  \ BIT-BANGING from host across serial line -- heavy delays
: sample-slow ( -- ) latch-in shiftin-byte show8 drop ;

\ Design: FAST - reading byte from shift-in register from target
\ Design: the uC takes care of the latching and reading using .AVR for 8-bit bit-banging.  
\ Design: Output stored in register so no realtime delay
\ Design: init-shiftin ENSURE to set control line mode appropriately: clock to OUTPUT and data to INPUT
\ Design: can use any bank of sequential registers (A, B, C) for DATA and PINS ( -- datareg  ctrlreg )
: bankA  24 28 ; \ for 13-key tone buttons, + 3 function buttons (octave up/down, X)
: bankB  35 32 ; \ for 24-bit switch array
: bankC  43 39 ; \ for 4x 16-key keypad + 1P4T
: shin8     ( datareg  ctrlreg -- byte )  over >r  2 nstore CALL_SHIFTINBYTE xcall r> [checkreg] nip ;
: shin16    ( datareg  ctrlreg -- bhi blo ) over >r 2 nstore CALL_SHIFTIN16  xcall r> checkreg ;
: shin24    ( datareg  ctrlreg -- bhi bmid blo ) over >r 2 nstore CALL_SHIFTIN24 xcall r> dup 1 + [checkreg] nip swap [checkreg] ;

: sample8  ( bankA/B -- b )            shin8 show8 ;  \ sample byte or 1/2 bar of beats (2x quarter notes, 8x 16th notes)
: sample16 ( bankA/B -- bhi blo )      shin16 show16 ;
: sample24 ( bankA/B -- bhi bmed blo ) shin24 show24 ;

\ Design: for use with 16-key KEYPAD + 1P4T module on bankC:
\ Design: encodings are: 16-31, 32-47, 64-79, 128-143,  for each of the 1P4T positions, from top to bottom
\ 1 byte (8-bit) output, 16 keys x 4
\ HARDWARE: ATMEGA328P uC + 1P4T switch + KS-60 4x4 matrix keypad + 74165 shiftin register on 400 & 120- tie point breadboards 5.50
\ : init-keymodule ( -- ) pinC-keypressed in ;
: check8  ( -- <encoding> )  pinC-keypressed pinread IF bankC shin8 ELSE 255 THEN ;  \ $FF if key not pressed
: 4bitID  ( nibble -- firstbitfromLSB1-4 ) 1 1  begin -rot 2dup = if  exitloop  else 1 lshift rot 1 + continueloop then until 2drop ;
: decode8 ( <encodedbyte not $FF> --  A-D #1-16 ) dup $F0 and 4 rshift 4bitID swap   $0F and 1 + ;
: poll8 ( -- A-D #1-16 ) begin check8 dup $FF < if decode8 exitloop else drop continueloop then until ;
: testpoll8 ( -- ) cr begin poll8 swap . . cr key? until ;  \ press button one more time after key to exit
\ SEE 3iMusicMenu.fs for use in ss [songselect]

\ HARDWARE LAYOUT of shift registers and switches.  Make the top switch position go into LSB.  Then 4bitID gives positions 1-4.

\ : Nsample  ( N -- N-bytes-lsb-TOS ) cr ." MSB" latch-in 0 do [sample] loop cr ." LSB" ; \ sample N bytes or N/2 bars of beats
\ : Nsample? ( N -- ) begin dup Nsample 2drop key? until drop ;

\ include 3iBeats.fs
\ : Nsamplebeat  ( N -- bytes N ) >r r@ Nsample r> ;  \ Nseq  --- add this if desired...


\ NOTE: don't forget decoupling capacitors on the IC
\ Also - remember need pull-down resistors on the buttons to ensure inputs aren't floating, and to prevent false readings.)
\ Pull down resistor at 10k on 5V supply will draw 1/2mA, so 8 drawing is 4mA.
\ Could use 100k to draw even less at 0.4mA.
\ 1 floating input is ok, but multiple and spurious readings occur

\ APPLICATIONS - see 3iBeats.fs

\ APPLICATION: - reading 1P4T switches through a shift-register
variable switchpos
: init-switchpos 0 switchpos c! ;
: switchpos++  switchpos c@ 1 + switchpos c! ;
\ Design: largest number is on top, e.g. on 1P4T 8 is up, 1 is down  (installing '165 with bottom facing up)
\ Design: converts shift-in reg read value into a switch position --- or error if not a clean read
: nibblepos ( 1hex -- pos/$FFerror ) init-switchpos 8 
    begin 2dup = 
        if exitloop 
        else switchpos++ 1 rshift dup 
            0 > if continueloop else $FF switchpos c! exitloop then
        then 
    until 2drop switchpos c@ ;
\ Design: parsing 1P4T with '165 installed bottom up, low nibble is on left and is first switch
: switch1 ( byte -- lo-pos ) $F and nibblepos ;
: bitswitch1 ( byte -- lo3 lo2 lo1 lo0 ) 4 0 do dup $1 and swap 1 rshift loop drop ;
: switch2 ( byte -- hi-pos ) $F0 and 4 rshift nibblepos ;
: 2nibbles ( byte -- hi-pos lo-pos ) dup switch2 swap switch1 ;
\ Design: parsing 4x 1P2T

\ 2x 1P4T switch values 0-indexed
: config1P4T ( -- hi-pos lo-pos ) bankB shin8 2nibbles ;
\ Design: reading a bank of 9 switches (4x 1P4T, 4x 1P2T, 1x 1P4T) and returning their individual 0-indexed values  CUSTOM LAYOUT
: config24  ( -- 9switchvalues ) bankB shin24  >r >r dup switch2 swap bitswitch1 r> 2nibbles r> 2nibbles ;
 


\ ************************
: help-shiftout
    cr 
    cr ." 74595 SHIFT OUT REGISTER"
    cr ." ------------------------"
    cr ."     use RealTerm to open the connection to the serial port"
    cr ." In GForth:"
    cr ."     include 3iForth, 3iShift.fs"
    cr ."     1 open-com"
    cr ."     init-shiftout"
    cr 
    cr ."   0 shout-byte    clr  (shifts right 8-bits)"
    cr
    cr ." Operating a shift out register:"
    cr ."    low-level:   pinA-data hi clk latch"
    cr ."    mid-level:   1/0 set-bit latch (pushes 0/1 into queue)"
    cr ."    high-level:  X set-id (1 pin, lsb is on RHS)  X set-byte (multi-pins)"
    cr ."    using embedded instructions (fastest): X shout-id   X shout-byte"
    cr ."    patterns:     kit  kit2  5-kit"
    cr
;


: help-shiftin
  cr 
  cr ." 74165 SHIFT IN REGISTER"
  cr ." -----------------------"
  cr ." include   3iForth.fs   3iBeats.fs  3iShift.fs (requires Nano running 3iForth.avr)"
  cr ." hardware: requires Nano, 74165 IC on nano pins 8,9,10, 8-bit dip or switches, buzzer on nano pin 4"
  cr ."   1 open-com"
  cr ."   init-shiftin"
  cr ."   sample-slow  [latch-in shiftin-byte show8]"
  cr
  cr ."   bankA (keyboard) | bankB (configuration)"
  cr ."   shin8 | shin16 | shin24"
  cr ."   bankA/B  sample8/16/24"
  cr ."   init-keymodule  check8 poll8"
  cr
  cr ."   Beats App:"
  cr ."      init-beats"
  cr ."      200 Nmetronome (quarter notes)  metronome (16th notes)"
  cr ."      [sample]   N Nsample  N Nbeat  N Nseq   N Nsamplebeat"
  cr ."      demo-uptown-funk"
  cr
;

\ ----------------------
help-shiftout
help-shiftin


\ V1 - 2021-04-10---Sat---20:52 --- Added ability to use POSI shift OUT register 74595
\ V2 - 2021-04-12---Mon---03:20 --- Added capability to use PISO shift IN register 74165
\ V3 - 2021-05-09---Sun---20:53 --- ADDED: words for piano32-lights board with 4 cascaded 74595 shift out registers
\       Successfully finished the Electrocookie based PCB prototype: 
\       Complexity: 32 leds, 32 resistors, 4x DIP-16 ICs, 5x capacitors, 6x pin header, 3-pin slide switch, 32 wire connections
\          74 components
\          275 connections hand-soldered connections
\       Defect: the power on switch doesn't really do anything, since power is run through the 4 control pins if the main power is off, it just leads to odd behaviour and probably stresses the chip.  What we probably need is for the keyboard to be electrically isolated from the microcontroller, i.e. we would want to use optoisolator between data pins on the micro and the board.
\ Next Steps:
\ See - 76489_speech_synth.txt   2021-04-09---Fri---04:36
\ V4 - 2021-05-17---Mon---05:11 -- ADDED: beats-sequencer from 8-bit SHIFT-IN register IC 74165s
\ V5 - 2021-05-18---Tue---02:22 -- ADDED: 16-beat beat sequencer from 2x cascaded shift IN register IC 74165
\ V6 - 2021-05-23---Sun---03:08 -- ADDED: software enabled beat sequencer 
\ V7 - 2021-05-24---Mon---15:29 -- ADDED: shin8 word that uses .avr function shiftinbyte for fast shift-in register usage
\ V8 - 2021-09-26---Sun---07:18 -- FEATURE: Added reading 1P4T switches and switch chord shp, with read error (ambiguous switch output) handling facility
\ V9 - 2021-09-27---Mon---00:45 -- FEATURE: Finished board with 24 HW switch inputs and added software to read the 24 switches using 3 shift-in registers
\ v10 - 2021-10-05---Tue---08:34 -- CAPABILITY: added ss [select song] which loads song from memory based on user input from 4x16 keypad using poll8